pub struct TDistribution { /* private fields */ }Expand description
Mixture distribution for Mandelstam t.
Implementations
Sourceimpl TDistribution
impl TDistribution
Sourcepub fn uniform() -> TDistribution
pub fn uniform() -> TDistribution
Construct a uniform distribution in t.
Sourcepub fn exponential(slope: f64) -> TDistribution
pub fn exponential(slope: f64) -> TDistribution
Construct an exponential distribution in t.
Sourcepub fn pole(exchange_mass: f64, power: f64) -> TDistribution
pub fn pole(exchange_mass: f64, power: f64) -> TDistribution
Construct a pole-like distribution in t.
Sourcepub fn histogram(histogram: Histogram) -> TDistribution
pub fn histogram(histogram: Histogram) -> TDistribution
Construct a histogram-backed distribution in t.
Sourcepub fn mixture(
components: impl IntoIterator<Item = (f64, TComponent)>,
) -> TDistribution
pub fn mixture( components: impl IntoIterator<Item = (f64, TComponent)>, ) -> TDistribution
Construct a weighted mixture of transfer-density components.
Sourcepub fn with_limits(
self,
t_min: Option<f64>,
t_max: Option<f64>,
) -> Result<TDistribution, LadduPhysicsError>
pub fn with_limits( self, t_min: Option<f64>, t_max: Option<f64>, ) -> Result<TDistribution, LadduPhysicsError>
Restrict this proposal to the intersection of these limits and the event-by-event physical t interval.
Errors
Returns LadduPhysicsError when a specified limit is non-finite or
t_min is not less than t_max.
Panics
Panics only if an option tested as present unexpectedly contains no value.