pub struct Parameter(/* private fields */);Expand description
An enum containing either a named free parameter or a constant value.
Implementations
Sourceimpl Parameter
impl Parameter
Sourcepub fn new(name: impl Into<String>) -> Parameter
pub fn new(name: impl Into<String>) -> Parameter
Create a free (floating) parameter with the given name.
Sourcepub fn new_fixed(name: impl Into<String>, value: f64) -> Parameter
pub fn new_fixed(name: impl Into<String>, value: f64) -> Parameter
Create a fixed parameter with the given name and value.
Sourcepub fn description(&self) -> Option<String>
pub fn description(&self) -> Option<String>
Return the optional human-readable description.
Sourcepub fn set_fixed_value(&self, value: Option<f64>)
pub fn set_fixed_value(&self, value: Option<f64>)
Helper method to set the fixed value of a parameter.
Sourcepub fn set_initial(&self, value: f64)
pub fn set_initial(&self, value: f64)
Helper method to set the initial value of a parameter.
Panics
This method panics if the parameter is fixed.
Sourcepub fn set_bounds<L, U>(&self, min: L, max: U)
pub fn set_bounds<L, U>(&self, min: L, max: U)
Helper method to set the bounds of a parameter.
Sourcepub fn set_latex(&self, latex: impl Into<String>)
pub fn set_latex(&self, latex: impl Into<String>)
Helper method to set the LaTeX representation of a parameter.
Sourcepub fn set_description(&self, description: impl Into<String>)
pub fn set_description(&self, description: impl Into<String>)
Helper method to set the description of a parameter.