Reaction

Struct Reaction 

pub struct Reaction { /* private fields */ }
A reaction with owned particle definitions and topology roles.

Implementations

Source

impl Reaction

pub fn new( graph: ParticleGraph, topology: ReactionTopology, ) -> Result<Reaction, LadduError>

Construct a reaction from a particle graph and topology.

pub fn two_to_two( p1: &Particle, p2: &Particle, p3: &Particle, p4: &Particle, ) -> Result<Reaction, LadduError>

Construct a two-to-two reaction.

pub const fn graph(&self) -> &ParticleGraph

Return the owned particle graph.

pub const fn topology(&self) -> &ReactionTopology

Return the reaction topology.

pub fn two_to_two_topology(&self) -> Result<&TwoToTwoReaction, LadduError>

Return the two-to-two topology role assignments.

pub fn role(&self, role: &str) -> Result<&Particle, LadduError>

Return the particle assigned to a named topology role.

pub fn p4( &self, event: &dyn EventLike, particle: &str, ) -> Result<Vec4, LadduError>

Resolve a particle p4 from an event.

pub fn resolve_two_to_two( &self, event: &dyn EventLike, ) -> Result<ResolvedTwoToTwo, LadduError>

Resolve the two-to-two topology momenta from an event.

pub fn mass(&self, particle: &str) -> Mass

Construct a mass variable for a particle.

pub fn decay(&self, parent: &str) -> Result<Decay, LadduError>

Construct an isobar decay view from a composite parent.

pub fn production(&self) -> Result<Production, LadduError>

Construct a two-to-two production view.

pub fn mandelstam(&self, channel: Channel) -> Result<Mandelstam, LadduError>

Construct a Mandelstam variable for this reaction.

pub fn pol_angle<A>(&self, angle_aux: A) -> PolAngle
where A: Into<String>,

Construct a polarization-angle variable for this reaction.

pub fn polarization<M, A>(&self, magnitude_aux: M, angle_aux: A) -> Polarization
where M: Into<String>, A: Into<String>,

Construct polarization variables for this reaction.

pub fn axes( &self, event: &dyn EventLike, particle: &str, frame: Frame, ) -> Result<FrameAxes, LadduError>

Compute axes for a particle in this reaction.

pub fn angles_value( &self, event: &dyn EventLike, parent: &str, daughter: &str, frame: Frame, ) -> Result<DecayAngles, LadduError>

Compute a daughter’s decay angles in its parent rest frame.

pub fn production_angles_value( &self, event: &dyn EventLike, produced: &str, frame: Frame, ) -> Result<DecayAngles, LadduError>

Compute the produced-system production angles in the overall center-of-momentum frame.

pub fn particle(&self, particle: &str) -> Result<&Particle, LadduError>

Return the particle with the given identifier.

pub fn contains(&self, particle: &str) -> bool

Return whether the reaction contains a particle with the given identifier.

pub fn particles(&self) -> Vec<&Particle>

Return all particle definitions in this reaction.