pub struct RejectionSamplingOptions {
pub target_accepted: usize,
pub generation_batch_size: usize,
pub output_batch_size: usize,
pub envelope: RejectionEnvelope,
pub seed: u64,
}Expand description
Options for rejection sampling generated events.
Fields
target_accepted: usizeNumber of accepted events to produce.
generation_batch_size: usizeNumber of raw events to generate per source batch.
output_batch_size: usizeTarget number of accepted events emitted per output batch.
envelope: RejectionEnvelopeEnvelope used by the rejection sampler.
seed: u64Random seed used for accept/reject decisions.