pub struct Voigt { /* private fields */ }Expand description
A Voigt line-shape Amplitude.
The normalized profile is
V(m; m_0, \Gamma, \sigma) =
\frac{\operatorname{Re}[w(z)]}{\sigma \sqrt{2\pi}},
\qquad
z = \frac{m - m_0 + i\Gamma/2}{\sigma\sqrt{2}},where $w(z)$ is the Faddeeva function. Here $\Gamma$ is the Lorentzian full
width at half maximum and $\sigma$ is the Gaussian resolution width. This amplitude
returns $\sqrt{V}$ so its norm-squared is the normalized Voigt profile.
Implementations
Sourceimpl Voigt
impl Voigt
Sourcepub fn new(
name: &str,
mass: Parameter,
width: Parameter,
sigma: Parameter,
resonance_mass: &Mass,
) -> Result<Expression, LadduError>
pub fn new( name: &str, mass: Parameter, width: Parameter, sigma: Parameter, resonance_mass: &Mass, ) -> Result<Expression, LadduError>
Construct a Voigt with the given name, mass, width, and Gaussian sigma.
This uses the given resonance_mass as the input mass. width is the Lorentzian full
width at half maximum and sigma is the Gaussian resolution width.